# -*- coding: utf-8 -*-
import base64
import hashlib
import json
import logging
import os
import re
from concurrent import futures
from pathlib import Path

import six
import xml

from ks3.bucketDecompressPolicy import BucketDecompressPolicy
from ks3.responseResult import ResponseResult, ResponseMetadata, IterableResponseResult
from ks3.share_encryption import ShareCryptor
from ks3.multipart import Part
from ks3.xmlParsers.bucketAccessMonitor import BucketAccessMonitor
from ks3.xmlParsers.bucketCors import BucketCors
from ks3.xmlParsers.bucketCrossReplicate import BucketCrossReplicate
from ks3.xmlParsers.bucketDataRedundancySwitch import BucketDataRedundancySwitch
from ks3.xmlParsers.bucketEncryption import BucketEncryption
from ks3.xmlParsers.bucketLifecycle import BucketLifecycle
from ks3.tagging import Tagging
from ks3.xmlParsers.bucketQos import BucketQos
from ks3.xmlParsers.bucketTransferAcceleration import BucketTransferAcceleration
from ks3.xmlParsers.requesterQos import RequesterQos

try:
    import urllib.parse as parse  # for Python 3
except ImportError:
    import urllib as parse  # for Python 2

from datetime import datetime as DT
from dateutil.tz import tzutc
from dateutil import parser
import ks3.utils
from ks3 import handler
from ks3.utils import RetryPolicy, ExponentialWait, StopAfterAttempt
from ks3.acl import Policy, CannedACLStrings
from ks3.bucketlistresultset import BucketListResultSet
from ks3.bucketlistresultset import MultiPartUploadListResultSet
from ks3.bucketlistresultset import VersionedBucketListResultSet
from ks3.bucketlistresultset import BucketRetentionListResultSet
from ks3.xmlParsers.bucketLogging import BucketLogging
from ks3.exception import S3ResponseError
from ks3.key import Key
from ks3.multipart import MultiPartUpload, CompleteMultiPartUpload
from ks3.prefix import Prefix
from ks3.resultset import ResultSet
from ks3.xmlParsers.bucketQuota import BucketQuota
from ks3.xmlParsers.bucketVersioning import BucketVersioningConfig
from ks3.deletemarker import DeleteMarker
from ks3.xmlParsers.bucketRetention import BucketRetention
from ks3.xmlParsers.bucketInventory import BucketInventory, ListInventoryConfigurationsResult

try:
    from ks3.encryption import Crypts
except:
    pass

logger = logging.getLogger(__name__)


# 检验conditions参数是否符合要求
def _check_conditions(conditions):
    allowed_types = {'eq', 'starts-with', 'content-length-range'}
    if conditions is not None:
        if not isinstance(conditions, list):
            raise ValueError("conditions必须是列表")
    bucket_flag = False
    for condition in conditions:
        if isinstance(condition, list):
            if len(condition) != 3:
                raise ValueError("列表格式的condition必须有3个元素")
            if condition[0] not in allowed_types:
                raise ValueError("condition的操作符必须是%s之一" % allowed_types)
            if condition[1] == '$bucket':
                bucket_flag = True
        if isinstance(condition, dict) and 'bucket' in condition:
            bucket_flag = True
    if not bucket_flag:
        raise ValueError("必须有包含bucket的condition")


def _build_complete_xml_from_parts(parts):
    """
    Build the XML body for complete_multipart_upload from a list of parts.

    :type parts: list or IterableResponseResult
    :param parts: List of Part objects, dictionaries, or IterableResponseResult
    :return: XML string
    """
    # 如果是 IterableResponseResult，获取其数据
    if isinstance(parts, IterableResponseResult):
        parts = parts.data

    xml_body = '<CompleteMultipartUpload>\n'
    for part in parts:
        xml_body += '  <Part>\n'

        # 支持 Part 对象
        if isinstance(part, Part):
            part_number = part.part_number
            etag = part.etag
        # 支持字典
        elif isinstance(part, dict):
            part_number = part.get('part_number') or part.get('PartNumber')
            etag = part.get('etag') or part.get('ETag')
            if part_number is None or etag is None:
                raise ValueError('每个 part 字典必须包含 "part_number" 和 "etag" 键')
        else:
            raise ValueError('Parts must be Part objects or dictionaries')

        # 添加验证
        if not isinstance(part_number, int) or part_number < 1:
          raise ValueError(f'part_number 必须是正整数，收到：{part_number}')
        if not etag or not isinstance(etag, str):
          raise ValueError(f'etag 必须是非空字符串，收到：{etag}')

        xml_body += '    <PartNumber>%d</PartNumber>\n' % int(part_number)
        xml_body += '    <ETag>%s</ETag>\n' % etag
        xml_body += '  </Part>\n'
    xml_body += '</CompleteMultipartUpload>'

    return xml_body


class Bucket(object):
    def __init__(self, connection=None, name=None, region=None, type=None, creation_date=None, *args, **kwargs):
        self.connection = connection
        self.name = name
        self.region = region
        self.type = type
        self.creation_date = creation_date

        self.response_metadata = ResponseMetadata(**kwargs)

    def __repr__(self):
        return '<Bucket: %s>' % self.name

    def __iter__(self):
        return iter(BucketListResultSet(self))

    def __contains__(self, key_name):
        return not (self.get_key(key_name, validate=True) is None)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Name':
            self.name = value
        elif name == 'CreationDate':
            self.creation_date = value
        elif name == 'Region':
            self.region = value
            setattr(self, name, value)  # 兼容旧的处理
        elif name == 'Type':
            self.type = value
            setattr(self, name, value)
        else:
            setattr(self, name, value)

    def new_key(self, key_name=None):
        """
        Creates a new key 
        
        :type key_name: string
        :param key_name: The name of the key to create

        :rtype: :class:`boto.s3.key.Key` or subclass
        :returns: An instance of the newly created key object
        """
        if not key_name:
            raise ValueError('Empty key names are not allowed')
        return Key(self, key_name)

    def copy_key(self, new_key_name, src_bucket_name, src_key_name, src_version_id=None, headers=None, query_args=None,
                 encrypt_key=False):
        """
        Create a new key in the bucket by copying another existing key.
        :param new_key_name: The name of the new key
        :param src_bucket_name: The name of the source bucket
        :param src_key_name: The name of the source key
        :param headers: A dictionary of header name/value pairs.
        :param query_args: A string of additional querystring arguments
            to append to the request
        :param encrypt_key: If True, the new copy of the object will
            be encrypted on the server-side by KS3 and will be stored
            in an encrypted form while at rest in KS3.
        :return:
        """
        if not new_key_name or not src_key_name:
            raise ValueError('Empty key names are not allowed')
        if not src_bucket_name:
            raise ValueError('Empty bucket name are not allowed')
        headers = headers or {}
        provider = self.connection.provider
        if encrypt_key:
            headers[provider.server_side_encryption_header] = 'AES256'
        src = '/%s/%s' % (src_bucket_name, parse.quote_plus(src_key_name.encode('utf-8')))
        src = src.replace('//', '/%2F')
        if src_version_id:
            src += '?versionId=%s' % src_version_id
        headers[provider.copy_source_header] = str(src)
        response = self.connection.make_request('PUT', self.name, new_key_name,
                                                headers=headers,
                                                query_args=query_args)
        body = response.read()
        if response.status == 200:
            key = self.new_key(new_key_name)
            key.handle_checksum_crc64ecma(response)
            h = handler.XmlHandler(key, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            if hasattr(key, 'Error'):
                raise provider.storage_copy_error(key.Code, key.Message, body)
            key.response_metadata = ResponseMetadata(status=response.status, reason=response.reason,
                                                     headers=response.headers)
            return key
        else:
            raise provider.storage_response_error(response.status,
                                                  response.reason, body)

    def generate_url(self, expires_in, method='GET', headers=None,
                     force_http=False, response_headers=None,
                     expires_in_absolute=False):
        return self.connection.generate_url(expires_in, method, self.name,
                                            headers=headers,
                                            force_http=force_http,
                                            response_headers=response_headers,
                                            expires_in_absolute=expires_in_absolute)

    def lookup(self, key_name, headers=None):
        k, resp = self._get_key_internal(key_name, headers, None)
        return k is not None

    # def delete_keys(self, keys, quiet=False, mfa_token=None, headers=None):
    #     """
    #     Deletes a set of keys using S3's Multi-object delete API. If a
    #     VersionID is specified for that key then that version is removed.
    #     Returns a MultiDeleteResult Object, which contains Deleted
    #     and Error elements for each key you ask to delete.
    #     """
    #     ikeys = iter(keys)
    #     result = MultiDeleteResult(self)
    #     provider = self.connection.provider
    #     query_args = 'delete'
    #
    #     def delete_keys2(hdrs):
    #         hdrs = hdrs or {}
    #         data = u"""<?xml version="1.0" encoding="UTF-8"?>"""
    #         data += u"<Delete>"
    #         if quiet:
    #             data += u"<Quiet>true</Quiet>"
    #         count = 0
    #         while count < 1000:
    #             try:
    #                 key = next(ikeys)
    #             except StopIteration:
    #                 break
    #             if isinstance(key, six.string_types):
    #                 key_name = key
    #                 version_id = None
    #             elif isinstance(key, tuple) and len(key) == 2:
    #                 key_name, version_id = key
    #             elif (isinstance(key, Key) or isinstance(key, DeleteMarker)) and key.name:
    #                 key_name = key.name
    #                 version_id = key.version_id
    #             else:
    #                 if isinstance(key, Prefix):
    #                     key_name = key.name
    #                     code = 'PrefixSkipped'   # Don't delete Prefix
    #                 else:
    #                     key_name = repr(key)   # try get a string
    #                     code = 'InvalidArgument'  # other unknown type
    #                 message = 'Invalid. No delete action taken for this object.'
    #                 error = Error(key_name, code=code, message=message)
    #                 result.errors.append(error)
    #                 continue
    #             count += 1
    #             data += u"<Object><Key>%s</Key>" % xml.sax.saxutils.escape(key_name)
    #             if version_id:
    #                 data += u"<VersionId>%s</VersionId>" % version_id
    #             data += u"</Object>"
    #         data += u"</Delete>"
    #         if count <= 0:
    #             return False  # no more
    #         data = data.encode('utf-8')
    #         fp = BytesIO(data)
    #         md5 = boto.utils.compute_md5(fp)
    #         hdrs['Content-MD5'] = md5[1]
    #         hdrs['Content-Type'] = 'text/xml'
    #         if mfa_token:
    #             hdrs[provider.mfa_header] = ' '.join(mfa_token)
    #         response = self.connection.make_request('POST', self.name,
    #                                                 headers=hdrs,
    #                                                 query_args=query_args,
    #                                                 data=data)
    #         body = response.read()
    #         if response.status == 200:
    #             h = handler.XmlHandler(result, self)
    #             if not isinstance(body, bytes):
    #                 body = body.encode('utf-8')
    #             xml.sax.parseString(body, h)
    #             return count >= 1000  # more?
    #         else:
    #             raise provider.storage_response_error(response.status,
    #                                                   response.reason,
    #                                                   body)
    #     while delete_keys2(headers):
    #         pass
    #     return result

    def get_key(self, key_name, headers=None, version_id=None,
                response_headers=None, validate=False):
        """
        Check to see if a particular key exists within the bucket.
        """
        if not validate:
            # if headers or version_id or response_headers:
            #     raise BotoClientError(
            #         "When providing 'validate=False', no other params " + \
            #         "are allowed."
            #     )

            # This leans on the default behavior of ``new_key`` (not hitting
            # the service). If that changes, that behavior should migrate here.
            return self.new_key(key_name)

        query_args_l = {}
        if version_id:
            query_args_l['versionId'] = version_id
        if response_headers:
            for rk, rv in six.iteritems(response_headers):
                query_args_l[rk] = parse.quote(rv)

        key, resp = self._get_key_internal(key_name, headers, query_args_l)
        return key

    def get_key_meta(self, key_name, headers=None, version_id=None,
                     response_headers=None, validate=True):
        """
        Check to see if a particular key exists within the bucket.
        """
        if validate is False:
            # if headers or version_id or response_headers:
            #     raise BotoClientError(
            #         "When providing 'validate=False', no other params " + \
            #         "are allowed."
            #     )

            # This leans on the default behavior of ``new_key`` (not hitting
            # the service). If that changes, that behavior should migrate here.
            return self.new_key(key_name)

        query_args_l = {}
        if version_id:
            query_args_l['versionId'] = version_id
        if response_headers:
            for rk, rv in six.iteritems(response_headers):
                query_args_l[rk] = parse.quote(rv)

        k, resp = self._get_key_internal(key_name, headers, query_args_l)
        return ResponseResult(data=resp, status=resp.status, reason=resp.reason,
                              headers=resp.headers)

    def _get_key_internal(self, key_name, headers, query_args_l):
        query_args = query_args_l or None
        response = self.connection.make_request('HEAD', self.name, key_name,
                                                headers=headers,
                                                query_args=query_args)
        response.read()
        # Allow any success status (2xx) - for example this lets us
        # support Range gets, which return status 206:
        if response.status // 100 == 2:
            k = Key(self)
            # provider = self.connection.provider
            # k.metadata = boto.utils.get_aws_metadata(response.msg, provider)
            for field in Key.base_fields:
                k.__dict__[field.lower().replace('-', '_')] = \
                    response.getheader(field)
            # the following machinations are a workaround to the fact that
            # apache/fastcgi omits the content-length header on HEAD
            # requests when the content-length is zero.
            # See http://goo.gl/0Tdax for more details.
            clen = response.getheader('content-length')
            if clen:
                k.size = int(response.getheader('content-length'))
            else:
                k.size = 0
            k.name = key_name
            k.handle_version_headers(response)
            k.handle_encryption_headers(response)
            k.handle_restore_headers(response)
            k.handle_addl_headers(response.getheaders())
            k.handle_user_metas(response)
            k.handle_storage_class(response)
            k.handle_tagging_count(response)
            k.handle_object_type(response)
            k.handle_next_position(response)
            k.handle_checksum_crc64ecma(response)
            k.response_metadata = ResponseMetadata(status=response.status, reason=response.reason,
                                                   headers=response.headers)
            return k, response
        else:
            if response.status == 404:
                return None, response
            else:
                raise S3ResponseError(response.status, response.reason, None,
                                      request_id=response.getheader('x-kss-request-id'))

    def list(self, prefix=None, delimiter='/', marker=None, max_keys=None,
             retry_policy=RetryPolicy(), request_interval=0):
        """
        List object keys within specified bucket.
        delimiter: str 默认为 '/' 。只会列举当前一层的文件和目录，且单次列举最多100个对象。
        如果设置为 delimiter='' 可以列举所有对象，且单次列举最多1000个对象。
        """
        return BucketListResultSet(self, prefix, delimiter, marker, max_keys,
                                   retry_policy=retry_policy, request_interval=request_interval)

    def list_retention(self, prefix=None, delimiter='/', marker=None, max_keys=None, request_interval=0):
        """
        delimiter: str 默认为 '/' 。只会列举当前一层的文件和目录，且单次列举最多100个对象。
        如果设置为 delimiter='' 可以列举所有对象，且单次列举最多1000个对象。
        """
        return BucketRetentionListResultSet(self, prefix, delimiter, marker, max_keys, request_interval=request_interval)

    def listObjects(self, prefix=None, delimiter=None, marker=None, max_keys=None,
                    filename=None, start_time=None, end_time=None, request_interval=0):
        keys = self.list(prefix, delimiter, marker, max_keys, request_interval=request_interval)
        keyLists = list()
        fd = None
        if filename is not None:
            fd = open(filename, "a")  # 利用追加模式,参数从w替换为a即可
            fd.seek(0)
        for k in keys:
            if isinstance(k, Key):
                if hasattr(DT, "timestamp"):
                    datetime = int(round(parser.parse(k.last_modified).timestamp()))
                else:
                    datetime = int(
                        round((parser.parse(k.last_modified) - DT(1970, 1, 1, tzinfo=tzutc())).total_seconds()))
                if start_time is None and end_time is not None:
                    if datetime >= end_time:
                        keyLists.append(k.name)
                        if fd:
                            fd.write("{}\n".format(k.name))
                elif end_time is None and start_time is not None:
                    if datetime <= start_time:
                        keyLists.append(k.name)
                        if fd:
                            fd.write("{}\n".format(k.name))
                elif end_time is not None and start_time is not None:
                    if start_time <= datetime or datetime <= end_time:
                        keyLists.append(k.name)
                        if fd:
                            fd.write("{}\n".format(k.name))
                elif end_time is None and start_time is None:
                    keyLists.append(k.name)
                    if fd:
                        fd.write("{}\n".format(k.name))

        if fd:
            fd.flush()
            fd.close()
        return keyLists

    def list_versions(self, prefix='', delimiter='', key_marker='',
                      version_id_marker='', headers=None, encoding_type=None,
                      request_interval=0):
        return VersionedBucketListResultSet(self, prefix, delimiter,
                                            key_marker, version_id_marker,
                                            headers,
                                            encoding_type=encoding_type,
                                            request_interval=request_interval)

    def list_v2(self, **params):
        """
        List object V2.
        """
        params_v2 = params.copy()
        params_v2['list_type'] = 2
        params_v2['delimiter'] = params_v2.get('delimiter', '/')
        return BucketListResultSet(self, **params_v2)

    def get_all_keys(self, retry_policy=None, **params):
        """
        A lower-level method for listing contents of a bucket.
        """
        return self._get_all([('Contents', Key), ('CommonPrefixes', Prefix)], '', retry_policy=retry_policy, **params)

    def get_all_versions(self, headers=None, **params):
        """
        A lower-level method for listing contents of a bucket.
        """
        return self._get_all([('Version', Key),
                              ('CommonPrefixes', Prefix),
                              ('DeleteMarker', DeleteMarker)],
                             'versions', headers, **params)

    def get_all_retention_keys(self, **params):
        return self._get_all([('Contents', Key), ('CommonPrefixes', Prefix)], initial_query_string='recycle', **params)

    def _get_all(self, element_map, initial_query_string='',
                 headers=None, retry_policy=None, **params):
        query_args = self._get_all_query_args(
            params,
            initial_query_string=initial_query_string
        )
        response = self.connection.make_request('GET', self.name,
                                                headers=headers,
                                                query_args=query_args,
                                                retry_policy=retry_policy)
        body = response.read()
        if response.status == 200:
            rs = ResultSet(element_map, status=response.status, reason=response.reason, headers=response.headers,
                           raw_body=body)
            h = handler.XmlHandler(rs, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return rs
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def _get_all_query_args(self, params, initial_query_string=''):
        pairs = []

        if initial_query_string:
            pairs.append(initial_query_string)

        for key, value in sorted(list(params.items()), key=lambda x: x[0]):
            if value is None:
                continue
            key = key.replace('_', '-')
            if key == 'maxkeys':
                key = 'max-keys'
            if not isinstance(value, six.string_types + (six.binary_type,)):
                value = six.text_type(value)
            if not isinstance(value, six.binary_type):
                value = value.encode('utf-8')
            if value:
                pairs.append('%s=%s' % (
                    parse.quote(key),
                    parse.quote(value)
                ))

        return '&'.join(pairs)

    def set_xml_acl(self, acl_str, key_name='', headers=None, version_id=None,
                    query_args=None):
        if query_args is None:
            query_args = {'acl': ''}
        if version_id:
            query_args['versionId'] = version_id
        if not isinstance(acl_str, bytes):
            acl_str = acl_str.encode('utf-8')
        if headers is None:
            headers = {}
        headers['content-type'] = 'application/xml'
        response = self.connection.make_request('PUT', self.name, key_name,
                                                data=acl_str,
                                                query_args=query_args,
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        if response.status != 200:
            raise S3ResponseError(response.status, response.reason, body)

    def set_acl(self, acl_or_str, key_name='', headers=None, version_id=None):
        if isinstance(acl_or_str, Policy):
            return self.set_xml_acl(acl_or_str.to_xml(), key_name,
                                    headers, version_id)
        else:
            return self.set_canned_acl(acl_or_str, key_name,
                                       headers, version_id)

    def set_canned_acl(self, acl_str, key_name='', headers=None,
                       version_id=None):
        assert acl_str in CannedACLStrings

        if headers:
            headers[self.connection.provider.acl_header] = acl_str
        else:
            headers = {self.connection.provider.acl_header: acl_str}

        query_args = 'acl'
        if version_id:
            query_args += '&versionId=%s' % version_id
        response = self.connection.make_request('PUT', self.name, key_name,
                                                headers=headers, query_args=query_args)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        elif response.status != 200:
            raise S3ResponseError(response.status, response.reason, body)

    def get_acl(self, key_name='', headers=None, version_id=None):
        query_args = 'acl'
        if version_id:
            query_args += '&versionId=%s' % version_id
        response = self.connection.make_request('GET', self.name, key_name,
                                                query_args=query_args,
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            policy = Policy(self, status=response.status, reason=response.reason, headers=response.headers,
                            raw_body=body)
            h = handler.XmlHandler(policy, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return policy
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def enable_logging(self, target_bucket, target_prefix='',
                       grants=None, headers=None):
        """
        Enable logging on a bucket.

        :param headers:
        :type target_bucket: bucket or string
        :param target_bucket: The bucket to log to.

        :type target_prefix: string
        :param target_prefix: The prefix which should be prepended to the
            generated log files written to the target_bucket.

        :type grants: list of Grant objects
        :param grants: A list of extra permissions which will be granted on
            the log files which are created.

        :rtype: bool
        :return: True if ok or raises an exception.
        """
        if isinstance(target_bucket, Bucket):
            target_bucket = target_bucket.name
        blogging = BucketLogging(target=target_bucket, target_prefix=target_prefix,
                                 grants=grants)
        return self.set_bucket_logging(blogging.to_xml(), headers=headers)

    def set_bucket_logging(self, logging_xml, headers=None):
        """
        Set logging on a bucket directly to the given xml string.

        :param headers:
        :type logging_xml: unicode string
        :param logging_xml: The XML for the bucketloggingstatus which
            will be set.  The string will be converted to utf-8 before
            it is sent.  Usually, you will obtain this XML from the
            BucketLogging object.

        :rtype: bool
        :return: True if ok or raises an exception.
        """

        if headers is None:
            headers = {}
        body = logging_xml
        if not isinstance(body, bytes):
            body = body.encode('utf-8')

        # md5 = ks3.utils.compute_base64_md5_digest(body)
        # headers['Content-MD5'] = md5
        headers['content-type'] = 'application/xml'
        response = self.connection.make_request('PUT', self.name, data=body,
                                                query_args='logging', headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def disable_logging(self, headers=None):
        """
        Disable logging on a bucket.

        :rtype: bool
        :return: True if ok or raises an exception.
        """
        blogging = BucketLogging()
        return self.set_bucket_logging(blogging.to_xml(), headers=headers)

    def get_bucket_logging(self, headers=None):
        """
        Get the logging for this bucket.

        :rtype: :class:`ks3.xmlParsers.bucketLogging.BucketLogging`
        :return: A BucketLogging object for this bucket.
        """
        response = self.connection.make_request('GET', self.name,
                                                query_args='logging', headers=headers)
        body = response.read()
        if response.status == 200:
            blogging = BucketLogging(status=response.status, reason=response.reason, headers=response.headers,
                                     raw_body=body)
            h = handler.XmlHandler(blogging, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return blogging
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_key(self, key_name, headers=None, version_id=None,
                   mfa_token=None):
        """
        Deletes a key from the bucket.
        """
        if not key_name:
            raise ValueError('Empty key names are not allowed')
        return self._delete_key_internal(key_name, headers=headers,
                                         version_id=version_id,
                                         mfa_token=mfa_token)

    def _delete_key_internal(self, key_name, headers=None, version_id=None,
                             mfa_token=None, query_args_l=None):
        query_args_l = query_args_l or ""
        provider = self.connection.provider
        if version_id:
            query_args_l.append('versionId=%s' % version_id)
        query_args = '&'.join(query_args_l) or None
        if mfa_token:
            if not headers:
                headers = {}
            headers[provider.mfa_header] = ' '.join(mfa_token)
        response = self.connection.make_request('DELETE', self.name, key_name,
                                                headers=headers,
                                                query_args=query_args)
        body = response.read()
        if response.status != 204:
            raise provider.storage_response_error(response.status,
                                                  response.reason, body)
        else:
            # return a key object with information on what was deleted.
            k = Key(self)
            k.name = key_name
            k.handle_version_headers(response)
            k.handle_addl_headers(response.getheaders())
            k.response_metadata = ResponseMetadata(status=response.status, reason=response.reason,
                                                   headers=response.headers)
            return k

    def list_multipart_uploads(self, key_marker='',
                               upload_id_marker='', prefix='', max_uploads=None, delimiter='',
                               headers=None, encoding_type=None, request_interval=0):
        """
        List multipart upload objects within a bucket.  This returns an
        instance of an MultiPartUploadListResultSet that automatically
        handles all of the result paging, etc. from S3.  You just need
        to keep iterating until there are no more results.

        :type key_marker: string
        :param key_marker: The "marker" of where you are in the result set

        :type upload_id_marker: string
        :param upload_id_marker: The upload identifier

        :param encoding_type: Requests Amazon S3 to encode the response and
            specifies the encoding method to use.

            An object key can contain any Unicode character; however, XML 1.0
            parser cannot parse some characters, such as characters with an
            ASCII value from 0 to 10. For characters that are not supported in
            XML 1.0, you can add this parameter to request that Amazon S3
            encode the keys in the response.

            Valid options: ``url``
        :type encoding_type: string

        :rtype: :class:`boto.s3.bucketlistresultset.BucketListResultSet`
        :return: an instance of a BucketListResultSet that handles paging, etc
        """
        return MultiPartUploadListResultSet(self, key_marker,
                                            upload_id_marker, prefix=prefix,
                                            max_uploads=max_uploads, delimiter=delimiter,
                                            headers=headers,
                                            encoding_type=encoding_type,
                                            request_interval=request_interval)

    def get_all_multipart_uploads(self, headers=None, **params):
        """
        A lower-level, version-aware method for listing active
        MultiPart uploads for a bucket.
        """
        # self.validate_kwarg_names(params, ['max_uploads', 'key_marker',
        #                                   'upload_id_marker', 'encoding_type',
        #                                   'delimiter', 'prefix'])
        return self._get_all([('Upload', MultiPartUpload),
                              ('CommonPrefixes', Prefix)],
                             'uploads', headers, **params)

    def initiate_multipart_upload(self, key_name, headers=None,
                                  reduced_redundancy=False,
                                  metadata=None, encrypt_key=False,
                                  policy=None, calc_encrypt_md5=True):
        """
        Start a multipart upload operation.
            Note: After you initiate multipart upload and upload one or more
            parts, you must either complete or abort multipart upload in order
            to stop getting charged for storage of the uploaded parts. Only
            after you either complete or abort multipart upload, Amazon S3
            frees up the parts storage and stops charging you for the parts
            storage.
        """
        query_args = 'uploads'
        provider = self.connection.provider
        headers = headers or {}
        if policy:
            headers[provider.acl_header] = policy
        if reduced_redundancy:
            storage_class_header = provider.storage_class_header
            if storage_class_header:
                headers[storage_class_header] = 'REDUCED_REDUNDANCY'
                # TODO: what if the provider doesn't support reduced redundancy?
        if encrypt_key:
            headers[provider.server_side_encryption_header] = 'AES256'
        if metadata is None:
            metadata = {}

        headers = ks3.utils.merge_meta(headers, metadata,
                                       self.connection.provider)
        if self.connection.local_encrypt:
            crypts = Crypts(self.connection.key)
            crypts.calc_md5 = calc_encrypt_md5
            crypts.action_info = "init_multi"
            md5_generator = hashlib.md5()
            md5_generator.update(crypts.key)
            headers["x-kss-meta-key"] = base64.b64encode(md5_generator.hexdigest().encode()).decode()
            headers["x-kss-meta-iv"] = base64.b64encode(crypts.first_iv).decode()
            response = self.connection.make_request('POST', self.name, key_name,
                                                    query_args=query_args,
                                                    headers=headers)
        else:
            response = self.connection.make_request('POST', self.name, key_name,
                                                    query_args=query_args,
                                                    headers=headers)
        body = response.read()
        if response.status == 200:
            resp = MultiPartUpload(self, status=response.status, reason=response.reason,
                                   headers=response.headers)
            if self.connection.local_encrypt:
                resp.set_crypt_context(crypts)
            h = handler.XmlHandler(resp, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return resp
        else:
            raise self.connection.provider.storage_response_error(
                response.status, response.reason, body)

    def list_parts(self, bucket_name, key_name, upload_id, max_parts=None,
                   part_number_marker=None, encoding_type=None):
        """
        List the parts that have been uploaded for a specific multipart upload.

        :type bucket_name: string
        :param bucket_name: The name of the bucket

        :type key_name: string
        :param key_name: The name of the key

        :type upload_id: string
        :param upload_id: The upload ID of the multipart upload

        :type max_parts: int
        :param max_parts: Maximum number of parts to return

        :type part_number_marker: int
        :param part_number_marker: Part number to start listing from

        :type encoding_type: string
        :param encoding_type: Encoding type for the response

        :rtype: :class:`ks3.responseResult.IterableResponseResult` or raises exception
        :returns: Iterable result containing the list of parts
        """
        query_args = 'uploadId=%s' % upload_id
        if max_parts:
            query_args += '&max-parts=%d' % max_parts
        if part_number_marker:
            query_args += '&part-number-marker=%s' % part_number_marker
        if encoding_type:
            query_args += '&encoding-type=%s' % encoding_type

        response = self.connection.make_request('GET', bucket_name, key_name,
                                              query_args=query_args)
        body = response.read()
        if response.status == 200:
            # Parse XML response to get parts list
            mp = MultiPartUpload(self, status=response.status, reason=response.reason,
                                headers=response.headers)
            mp._parts = []
            mp.bucket_name = bucket_name
            mp.key_name = key_name
            mp.id = upload_id
            mp.part_number_marker = part_number_marker

            h = handler.XmlHandler(mp, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            
            try:
                xml.sax.parseString(body, h)
            except xml.sax.SAXParseException as e:
                logger.error('XML parse error at line %d, column %d: %s', e.getLineNumber(), e.getColumnNumber(), e.getMessage())
                logger.error('Response body: %s', body.decode('utf-8') if isinstance(body, bytes) else body)
                raise

            return IterableResponseResult(mp._parts, status=response.status, reason=response.reason,
                                        headers=response.headers)
        else:
            raise self.connection.provider.storage_response_error(
                response.status, response.reason, body)

    def upload_part_from_file(self, key_name, upload_id, fp, part_num,
                              headers=None, replace=True, cb=None, num_cb=10,
                              md5=None, size=None, crypt_context=None,
                              is_last_part="origin"):
        """
        Upload a part of a multipart upload from a file.

        :type key_name: string
        :param key_name: 待上传文件名，要和 `initiate_multipart_upload` 的文件名一致。

        :type upload_id: string
        :param upload_id: The upload ID of the multipart upload

        :type fp: file
        :param fp: The file pointer to upload from

        :type part_num: int
        :param part_num: The part number (must be > 0)

        :type headers: dict
        :param headers: Headers to send with the request

        :type replace: bool
        :param replace: Whether to replace existing data

        :type cb: function
        :param cb: Callback function for progress

        :type num_cb: int
        :param num_cb: Number of callbacks

        :type md5: tuple
        :param md5: MD5 hash tuple

        :type size: int
        :param size: Size of the data to upload

        :type crypt_context: Crypts
        :param crypt_context: Encryption context for local encryption

        :type is_last_part: str or bool
        :param is_last_part: Whether this is the last part

        :rtype: :class:`httplib.HTTPResponse`
        :returns: The upload response
        """
        if part_num < 1:
            raise ValueError('Part numbers must be greater than zero')

        query_args = 'partNumber=%d&uploadId=%s' % (part_num, upload_id)
        key = self.new_key(key_name)

        # Handle local encryption
        if self.connection.local_encrypt and crypt_context:
            # Get sizes of the whole file and the part
            fp.seek(0, os.SEEK_END)
            part_size = fp.tell()
            fp.seek(0, os.SEEK_SET)
            assert part_size, "upload part size can not be 0 !"

            if not is_last_part:
                assert (part_size % 16 == 0), \
                    "The part size must be multiples of 16 except the last part in local encrypt mode."

            crypt_context.part_num = part_num
            crypt_context.is_last_part = is_last_part
            crypt_context.action_info = "upload_part"
            response = key.set_contents_from_file(fp, headers=headers, replace=replace,
                                                  cb=cb, num_cb=num_cb, md5=md5,
                                                  reduced_redundancy=False,
                                                  query_args=query_args, size=size,
                                                  crypt_context=crypt_context)
        else:
            response = key.set_contents_from_file(fp, headers=headers, replace=replace,
                                                  cb=cb, num_cb=num_cb, md5=md5,
                                                  reduced_redundancy=False,
                                                  query_args=query_args, size=size)

        return response

    def copy_part_from_key(self, key_name, upload_id, src_bucket_name, 
                           src_key_name, part_num, start=None, end=None,
                           headers=None):
        """
        Copy a part of a multipart upload from an existing key.

        :type key_name: string
        :param key_name: 待上传文件名，要和 `initiate_multipart_upload` 的文件名一致。

        :type upload_id: string
        :param upload_id: The upload ID of the multipart upload

        :type src_bucket_name: string
        :param src_bucket_name: Name of the bucket containing the source key

        :type src_key_name: string
        :param src_key_name: Name of the source key

        :type part_num: int
        :param part_num: The part number (must be > 0)

        :type start: int
        :param start: Zero-based byte offset to start copying from

        :type end: int
        :param end: Zero-based byte offset to copy to

        :type headers: dict
        :param headers: Any headers to pass along in the request

        :rtype: :class:`ks3.key.Key`
        :returns: The copied part containing the etag
        """
        if part_num < 1:
            raise ValueError('Part numbers must be greater than zero')
        
        query_args = 'uploadId=%s&partNumber=%d' % (upload_id, part_num)
        
        if start is not None and end is not None:
            rng = 'bytes=%s-%s' % (start, end)
            provider = self.connection.provider
            if headers is None:
                headers = {}
            else:
                headers = headers.copy()
            headers[provider.copy_source_range_header] = rng
        
        return self.copy_key(key_name, src_bucket_name,
                            src_key_name,
                            headers=headers,
                            query_args=query_args)

    def complete_multipart_upload(self, key_name, upload_id,
                                  xml_body=None, headers=None, parts=None):
        """
        Complete a multipart upload operation.

        :type key_name: string
        :param key_name: 待上传文件名，要和 `initiate_multipart_upload` 的文件名一致。

        :type upload_id: string
        :param upload_id: The upload ID of the multipart upload

        :type xml_body: string
        :param xml_body: The XML body containing the list of parts (deprecated, use parts parameter instead)

        :type headers: dict
        :param headers: Headers to send with the request

        :type parts: list or IterableResponseResult
        :param parts: List of Part objects or dictionaries with 'part_number' and 'etag' keys,
                     or IterableResponseResult from list_parts method.
                     If provided, xml_body will be automatically generated.
        """
        # 如果提供了 parts 参数，则自动生成 XML
        if parts is not None:
            if xml_body is not None:
                raise ValueError('Cannot specify both xml_body and parts parameters')
            xml_body = _build_complete_xml_from_parts(parts)
        elif xml_body is None:
          raise ValueError('必须提供 xml_body 或 parts 参数之一。'
                           '建议使用 parts 参数传入 Part 对象列表或 list_parts生成的IterableResponseResult。')

        query_args = 'uploadId=%s' % upload_id
        if headers is None:
            headers = {}
        headers['Content-Type'] = 'text/xml'
        logger.debug('key: {0}, bucket: {1}, upload_id: {2}, request_body: {3}'
                     .format(key_name, self.name, upload_id, xml_body))
        response = self.connection.make_request('POST', self.name, key_name,
                                                query_args=query_args,
                                                headers=headers, data=xml_body)
        contains_error = False
        body = response.read().decode('utf-8')
        # Some errors will be reported in the body of the response
        # even though the HTTP response code is 200.  This check
        # does a quick and dirty peek in the body for an error element.
        if body.find('<Error>') > 0:
            contains_error = True
        if response.status == 200 and not contains_error:
            resp = CompleteMultiPartUpload(self, status=response.status, reason=response.reason,
                                           headers=response.headers)
            h = handler.XmlHandler(resp, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            # Use a dummy key to parse various response headers
            # for versioning, encryption info and then explicitly
            # set the completed MPU object values from key.
            k = Key(self)
            k.handle_version_headers(response)
            k.handle_encryption_headers(response)
            resp.version_id = k.version_id
            resp.encrypted = k.encrypted
            resp.status = response.status
            return resp
        else:
            raise self.connection.provider.storage_response_error(
                response.status, response.reason, body)

    def cancel_multipart_upload(self, key_name, upload_id, headers=None):
        """
        To verify that all parts have been removed, so you don't get charged
        for the part storage, you should call the List Parts operation and
        ensure the parts list is empty.
        """
        query_args = 'uploadId=%s' % upload_id
        response = self.connection.make_request('DELETE', self.name, key_name,
                                                query_args=query_args,
                                                headers=headers)
        body = response.read()
        if response.status != 204:
            raise self.connection.provider.storage_response_error(
                response.status, response.reason, body)
        return ResponseResult(data=response, status=response.status, reason=response.reason,
                              headers=response.headers)

    def set_adp(self, key_name, headers):
        query_args = 'adp'
        response = self.connection.make_request('PUT', self.name, key_name,
                                                headers=headers, query_args=query_args)
        body = response.read()
        if response.status != 200:
            raise S3ResponseError(response.status, response.reason, body)
        task_id = response.getheader("TaskID")
        return ResponseResult(task_id, status=response.status, reason=response.reason,
                              headers=response.headers)

    def set_bucket_quota(self, headers=None, quota=0):
        """
        :param quota: bucket quota default 0 not limit
        :param headers: custom header
        :return: True is ok or raises an exception.
        """
        bucketQuota = BucketQuota(quota)
        quota_to_xml = bucketQuota.to_xml()
        if not isinstance(quota_to_xml, bytes):
            quota_to_xml = quota_to_xml.encode('utf-8')
        if headers is None:
            headers = {}
        headers['Content-Type'] = 'text/xml'
        response = self.connection.make_request('PUT', self.name, data=quota_to_xml,
                                                query_args='quota', headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_quota(self, headers=None):
        response = self.connection.make_request('GET', self.name,
                                                query_args='quota', headers=headers)
        body = response.read()
        if response.status == 200:
            quota = BucketQuota()
            h = handler.XmlHandler(quota, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return quota
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_policy(self, policy, headers=None):
        logger.debug('bucket: {0}, bucket_policy request_body: {1}'.format(self.name, policy))
        response = self.connection.make_request('PUT', self.name, data=policy,
                                                query_args='policy', headers=headers)
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_policy(self, headers=None):
        response = self.connection.make_request('GET', self.name,
                                                query_args='policy', headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(body, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_policy(self, headers=None):
        response = self.connection.make_request('DELETE', self.name,
                                                query_args='policy', headers=headers)
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_crr(self, target_bucket, delete_marker_status=BucketCrossReplicate.DISABLED, prefix_list=None,
                       historical_object_replication=BucketCrossReplicate.DISABLED, headers=None):
        replicate = BucketCrossReplicate(target_bucket, delete_marker_status, prefix_list,
                                         historical_object_replication)
        if headers is None:
            headers = {}
        headers['Content-Type'] = 'text/xml'
        replicate_xml = replicate.to_xml()
        logger.debug('bucket: {0}, bucket_crr request_body: {1}'.format(self.name, replicate_xml))
        if not isinstance(replicate_xml, bytes):
            replicate_xml = replicate_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(replicate_xml)
        headers['Content-MD5'] = md5
        response = self.connection.make_request('PUT', self.name, data=replicate_xml,
                                                query_args='crr', headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_crr(self, headers=None):
        response = self.connection.make_request('GET', self.name,
                                                query_args='crr', headers=headers)
        body = response.read()
        if response.status == 200:
            replicate = BucketCrossReplicate(status=response.status, reason=response.reason, headers=response.headers,
                                             raw_body=body)
            h = handler.XmlHandler(replicate, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return replicate
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_crr(self, headers=None):
        response = self.connection.make_request('DELETE', self.name,
                                                query_args='crr', headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    # deprecated. moved to key.py
    def restore_object(self, object_key_name, days=None, tier=None, headers=None):
        """
        Restore an archived object.

        :type object_key_name: str
        :param object_key_name: The name of the object to restore

        :type days: int
        :param days: Number of days to keep the restored object available

        :type tier: str
        :param tier: Retrieval tier (e.g., 'Standard', 'Expedited', 'Bulk')

        :type headers: dict
        :param headers: Additional HTTP headers

        :rtype: ResponseResult
        :return: Response result
        """
        # Build the restore request XML
        if days is not None or tier is not None:
            restore_body = '<RestoreRequest>\n'
            if days is not None:
                restore_body += '    <Days>{0}</Days>\n'.format(days)
            if tier is not None:
                restore_body += '    <JobParameters>\n'
                restore_body += '        <Tier>{0}</Tier>\n'.format(tier)
                restore_body += '    </JobParameters>\n'
            restore_body += '</RestoreRequest>'
        else:
            restore_body = ''

        response = self.connection.make_request('POST', self.name, object_key_name,
                                                data=restore_body,
                                                query_args='restore', headers=headers)
        body = response.read()
        if response.status == 200 or response.status == 202:
            return ResponseResult(body, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    # deprecated. moved to key.py
    def set_object_tagging(self, object_key_name, tagging_set, version_id=None, headers=None):
        object_tagging = Tagging(tagging_set)
        if headers is None:
            headers = {}
        headers['Content-Type'] = 'text/xml'
        object_tagging_xml = object_tagging.to_xml()
        if not isinstance(object_tagging_xml, bytes):
            object_tagging_xml = object_tagging_xml.encode('utf-8')

        query_args = 'tagging'
        if version_id is not None:
            query_args = query_args + '&versionId=' + version_id
        logger.debug('bucket: {0}, key: {1}, request_body: {2}'.format(self.name, object_key_name, object_tagging_xml))
        response = self.connection.make_request('PUT', self.name, object_key_name, data=object_tagging_xml,
                                                query_args=query_args, headers=headers)

        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    # deprecated. moved to key.py
    def get_object_tagging(self, object_key_name, version_id=None, headers=None):

        query_args = 'tagging'
        if version_id is not None:
            query_args = query_args + '&versionId=' + version_id
        response = self.connection.make_request('GET', self.name, object_key_name,
                                                query_args=query_args, headers=headers)
        body = response.read()
        if response.status == 200:
            objectTagging = Tagging()
            h = handler.XmlHandler(objectTagging, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return objectTagging
        else:
            raise S3ResponseError(response.status, response.reason, body)

    # deprecated. moved to key.py
    def delete_object_tagging(self, object_key_name, version_id=None, headers=None):
        query_args = 'tagging'
        if version_id is not None:
            query_args = query_args + '&versionId=' + version_id
        response = self.connection.make_request('DELETE', self.name, object_key_name,
                                                query_args=query_args, headers=headers)
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    # deprecated. moved to key.py
    def fetch_object(self, object_key_name, source_url=None, callback_url=None, headers=None):
        if headers is None:
            headers = {}
        if source_url is not None:
            headers['x-kss-sourceurl'] = source_url
        if callback_url is not None:
            headers['x-kss-callbackurl'] = callback_url

        response = self.connection.make_request('PUT', self.name, object_key_name,
                                                query_args='fetch', headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_lifecycle(self, bucketLifecycle, headers=None):
        if headers is None:
            headers = {}
        bucketLifecycle_xml = bucketLifecycle.to_xml()
        if not isinstance(bucketLifecycle_xml, bytes):
            bucketLifecycle_xml = bucketLifecycle_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(bucketLifecycle_xml)
        headers['Content-MD5'] = md5
        headers['content-type'] = 'application/xml'
        logger.debug('bucket: {0}, bucket_lifecycle request_body: {1}'.format(self.name, bucketLifecycle_xml))
        response = self.connection.make_request('PUT', self.name, data=bucketLifecycle_xml,
                                                query_args='lifecycle', headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_lifecycle(self, headers=None):
        response = self.connection.make_request('GET', self.name,
                                                query_args='lifecycle', headers=headers)
        body = response.read()
        if response.status == 200:
            lifecycle = BucketLifecycle(status=response.status, reason=response.reason, headers=response.headers,
                                        raw_body=body)
            h = handler.XmlHandler(lifecycle, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return lifecycle
        else:
            raise S3ResponseError(response.status, response.reason, body)

    # delete bucket lifecycle
    def delete_bucket_lifecycle(self, headers=None):
        response = self.connection.make_request('DELETE', self.name,
                                                query_args='lifecycle', headers=headers)
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_versioning(self, version_configuration, headers=None):
        if headers is None:
            headers = {}
        version_configuration_xml = version_configuration.to_xml()
        if not isinstance(version_configuration_xml, bytes):
            version_configuration_xml = version_configuration_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(version_configuration_xml)
        headers['Content-MD5'] = md5
        headers['content-type'] = 'application/xml'
        logger.debug('bucket: {0}, bucket_versioning request_body: {1}'.format(self.name, version_configuration_xml))
        response = self.connection.make_request('PUT', self.name, data=version_configuration_xml,
                                                query_args='versioning',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_versioning(self, headers=None):
        response = self.connection.make_request('GET', self.name,
                                                query_args='versioning', headers=headers)
        body = response.read()
        if response.status == 200:
            versioningConfig = BucketVersioningConfig(status=response.status, reason=response.reason,
                                                      headers=response.headers, raw_body=body)
            h = handler.XmlHandler(versioningConfig, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return versioningConfig
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_cors(self):
        response = self.connection.make_request('GET', self.name,
                                                query_args='cors')
        body = response.read()
        if response.status == 200:
            bucketCors = BucketCors(status=response.status, reason=response.reason, headers=response.headers,
                                    raw_body=body)
            h = handler.XmlHandler(bucketCors, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return bucketCors
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_cors(self, bucket_cors):
        headers = {'content-type': 'application/xml'}
        bucket_cors_xml = bucket_cors.to_xml()
        if not isinstance(bucket_cors_xml, bytes):
            bucket_cors_xml = bucket_cors_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(bucket_cors_xml)
        headers['content-md5'] = md5
        logger.debug('bucket: {0}, bucket_cors request_body: {1}'.format(self.name, bucket_cors_xml))
        response = self.connection.make_request('PUT', self.name, data=bucket_cors_xml,
                                                query_args='cors',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_cors(self):
        response = self.connection.make_request('DELETE', self.name, query_args='cors')
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_mirror(self):
        response = self.connection.make_request('GET', self.name, query_args='mirror')
        body = response.read()
        if response.status == 200:
            return ResponseResult(body, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_mirror(self, mirror, headers=None):
        headers = {'content-type': 'application/json'}
        if headers is None:
            headers = {}
        mirror_dict = dict(mirror)
        mirror_json = json.dumps(mirror_dict)
        if not isinstance(mirror_json, bytes):
            mirror_json = mirror_json.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(mirror_json)
        headers['content-md5'] = md5
        logger.debug('bucket: {0}, bucket_mirror request_body: {1}'.format(self.name, mirror_json))
        response = self.connection.make_request('PUT', self.name, data=mirror_json,
                                                query_args='mirror',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_mirror(self):
        response = self.connection.make_request('DELETE', self.name, query_args='mirror')
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_encryption(self):
        response = self.connection.make_request('DELETE', self.name, query_args='encryption')
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_retention(self, bucket_retention):
        headers = {'content-type': 'application/xml'}
        retention_xml = bucket_retention.to_xml()
        if not isinstance(retention_xml, bytes):
            retention_xml = retention_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(retention_xml)
        headers['Content-MD5'] = md5
        logger.debug('bucket: {0}, bucket_retention request_body: {1}'.format(self.name, retention_xml))
        response = self.connection.make_request('PUT', self.name, data=retention_xml,
                                                query_args='retention',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_encryption(self, bucket_encryption):
        headers = {'content-type': 'application/xml'}
        encryption_xml = bucket_encryption.to_xml()
        if not isinstance(encryption_xml, bytes):
            encryption_xml = encryption_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(encryption_xml)
        headers['Content-MD5'] = md5
        logger.debug('bucket: {0}, bucket_encryption request_body: {1}'.format(self.name, encryption_xml))
        response = self.connection.make_request('PUT', self.name, data=encryption_xml,
                                                query_args='encryption',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_access_monitor(self, bucket_access_monitor):
        headers = {'content-type': 'application/xml'}
        ac_xml = bucket_access_monitor.to_xml()
        if not isinstance(ac_xml, bytes):
            ac_xml = ac_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(ac_xml)
        headers['Content-MD5'] = md5
        logger.debug('bucket: {0}, bucket_access_monitor request_body: {1}'.format(self.name, ac_xml))
        response = self.connection.make_request('PUT', self.name, data=ac_xml,
                                                query_args='accessmonitor',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_retention(self):
        response = self.connection.make_request('GET', self.name, query_args='retention')
        body = response.read()
        if response.status == 200:
            retention = BucketRetention(status=response.status, reason=response.reason, headers=response.headers,
                                        raw_body=body)
            h = handler.XmlHandler(retention, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return retention
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_encryption(self):
        response = self.connection.make_request('GET', self.name, query_args='encryption')
        body = response.read()
        if response.status == 200:
            encryption = BucketEncryption(status=response.status, reason=response.reason, headers=response.headers,
                                          raw_body=body)
            h = handler.XmlHandler(encryption, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return encryption
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_access_monitor(self):
        response = self.connection.make_request('GET', self.name, query_args='accessmonitor')
        body = response.read()
        if response.status == 200:
            access_monitor = BucketAccessMonitor(status=response.status, reason=response.reason, headers=response.headers,
                                          raw_body=body)
            h = handler.XmlHandler(access_monitor, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return access_monitor
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_inventory(self, bucket_inventory):
        headers = {'content-type': 'application/xml'}
        inventory_xml = bucket_inventory.to_xml()
        if not isinstance(inventory_xml, bytes):
            inventory_xml = inventory_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(inventory_xml)
        headers['Content-MD5'] = md5
        logger.debug('bucket: {0}, bucket_inventory request_body: {1}'.format(self.name, inventory_xml))
        response = self.connection.make_request('PUT', self.name, data=inventory_xml,
                                                query_args='inventory&id=%s' % bucket_inventory.id,
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_inventory(self, inventory_id):
        response = self.connection.make_request('GET', self.name, query_args='inventory&id=%s' % inventory_id)
        body = response.read()
        if response.status == 200:
            inventory = BucketInventory(status=response.status, reason=response.reason, headers=response.headers,
                                        raw_body=body)
            h = handler.XmlHandler(inventory, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return inventory
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def list_bucket_inventory(self):
        response = self.connection.make_request('GET', self.name, query_args='inventory')
        body = response.read()
        if response.status == 200:
            result = ListInventoryConfigurationsResult(status=response.status, reason=response.reason,
                                                       headers=response.headers,
                                                       raw_body=body)
            h = handler.XmlHandler(result, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return result
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_inventory(self, inventory_id):
        response = self.connection.make_request('DELETE', self.name, query_args='inventory&id=%s' % inventory_id)
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def upload_dir(
            self, directory, key_prefix='',
            upload_threads=3,
            part_upload_threads=3,
            part_size=100 * 1024,
            headers=None,
    ):
        dir_path = Path(directory)
        executor = ks3.utils.BlockThreadPoolExecutor(
            max_workers=upload_threads,
            thread_name_prefix='upload-dir-task',
        )
        file_futures = []
        for file in dir_path.rglob('*'):
            if file.is_dir():
                continue
            key = self.new_key(key_prefix + file.relative_to(dir_path).as_posix())
            filename = file.as_posix()
            future = executor.submit(
                key.upload_file,
                filename,
                part_size=part_size,
                threads_num=part_upload_threads,
                headers=headers,
            )
            file_futures.append(future)
        executor.shutdown(wait=True)
        for future in futures.as_completed(file_futures):
            future.result()

    def set_bucket_tagging(self, tagging_set):
        headers = {'content-type': 'text/xml'}
        bucket_tagging = Tagging(tagging_set)
        tagging_xml = bucket_tagging.to_xml()
        if not isinstance(tagging_xml, bytes):
            tagging_xml = tagging_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(tagging_xml)
        headers['Content-MD5'] = md5
        logger.debug('bucket: {0}, bucket_tagging request_body: {1}'.format(self.name, tagging_xml))
        response = self.connection.make_request('PUT', self.name, data=tagging_xml,
                                                query_args='tagging',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_tagging(self):
        response = self.connection.make_request('GET', self.name, query_args='tagging')
        body = response.read()
        if response.status == 200:
            tagging = Tagging(status=response.status, reason=response.reason, headers=response.headers,
                                          raw_body=body)
            h = handler.XmlHandler(tagging, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return tagging
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_tagging(self):
        response = self.connection.make_request('DELETE', self.name, query_args='tagging')
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_decompress_policy(self, decompress_policy: BucketDecompressPolicy=None):
        decompress_policy_raw = decompress_policy.to_json()
        return self.set_bucket_decompress_policy_raw(decompress_policy_raw)

    def set_bucket_decompress_policy_raw(self, decompress_policy_raw):
        headers = {'content-type': 'application/json'}
        logger.debug('bucket: {0}, decompress_policy request_body: {1}'.format(self.name, decompress_policy_raw))
        response = self.connection.make_request('PUT', self.name, data=decompress_policy_raw,
                                                query_args='decompresspolicy',
                                                headers=headers)
        body = response.read()
        if 200 <= response.status < 300:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_decompress_policy(self):
        response = self.connection.make_request('GET', self.name, query_args='decompresspolicy')
        body = response.read()
        if 200 <= response.status < 300:
            policy = BucketDecompressPolicy.from_json(body)
            response_metadata = ResponseMetadata(status=response.status, reason=response.reason, headers=response.headers)
            policy.response_metadata = response_metadata
            return policy
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_decompress_policy(self):
        response = self.connection.make_request('DELETE', self.name, query_args='decompresspolicy')
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)
    # 调整名字为 presign，以区分key里的get_presigned_url
    def presign(self, expires_in=3600, method="PUT", key_name=None, conditions=None, headers=None,
                          params=None, expires_in_absolute=False, version_id=None, domain=False):
        """
        生成临时上传链接
        :param conditions: policy 条件组
        :param expires_in: 过期时间 默认为 3600秒
        """
        if conditions is not None:
            _check_conditions(conditions)
        if headers is None:
            headers = {}
        else:
            headers = headers.copy()

        url = self.connection.generate_url(expires_in, method, bucket=self.name, key=key_name,
                                           headers=headers, params=params,
                                           expires_in_absolute=expires_in_absolute, version_id=version_id,
                                           conditions=conditions)
        if domain is True:
            url = url.replace('http://' + self.name + '.', 'http://')
            url = url.replace('https://' + self.name + '.', 'https://')

        return url

    def create_share_code(self, auth_code=None, expires_in=86400, prefix=None):
        """
        生成授权码的临时访问token
        :param auth_code: 授权码，6位数字或者字母的组合
        :param expires_in: 过期时间 默认为 1天
        :param prefix: 要分享的前缀
        """
        conditions = [{'bucket': self.name}]

        if auth_code is None:
            raise ValueError('auth_code is required')

        # auth_code应该是字母或者数字的组合，用正则检查
        if not re.match(rb'^[a-zA-Z0-9]{6}$', auth_code):
            raise ValueError('auth_code应该是6位数字或者字母的组合')

        list_params = {}
        if prefix is not None:
            conditions.append(['starts-with', '$key', prefix])
            list_params['prefix'] = prefix

        share_url = self.presign(expires_in, method='GET', params=list_params, conditions=conditions)
        logger.debug('share_url: {0}'.format(share_url))
        shareCryptor = ShareCryptor(auth_code)
        # salt = os.urandom(16)
        share_code = shareCryptor.encrypt(share_url)
        logger.debug('share_code: {0}'.format(share_code))
        return share_code

    def create_share_url(self, auth_code=None, expires_in=86400, prefix=None,
                         share_url_base='https://ks3.console.ksyun.com/doc-share.html?token='):
        """
        生成授权码的临时访问页面链接
        :param auth_code: 授权码
        :param expires_in: 过期时间 默认为 1天
        :param prefix: 要分享的前缀
        :param share_url_base: 分享页的承载地址
        """
        if isinstance(auth_code, int):
            auth_code = str(auth_code)
        if not isinstance(auth_code, bytes):
            auth_code = auth_code.encode('utf-8')
        share_code = self.create_share_code(auth_code=auth_code, expires_in=expires_in,
                                            prefix=prefix)
        share_url = share_url_base + share_code
        return share_url

    def set_bucket_qos(self, bucket_qos):
        headers = {'content-type': 'application/xml'}
        bucket_qos_xml = bucket_qos.to_xml()
        if not isinstance(bucket_qos_xml, bytes):
            bucket_qos_xml = bucket_qos_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(bucket_qos_xml)
        headers['Content-MD5'] = md5
        logger.debug('bucket: {0}, request_body: {1}'.format(self.name, bucket_qos_xml))
        response = self.connection.make_request('PUT', self.name, data=bucket_qos_xml, query_args='bucketqos',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_qos(self):
        response = self.connection.make_request('GET', self.name, query_args='bucketqos')
        body = response.read()
        if response.status == 200:
            bucket_quota = BucketQos(status=response.status, reason=response.reason, headers=response.headers,
                                     raw_body=body)
            h = handler.XmlHandler(bucket_quota, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return bucket_quota
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket_qos(self):
        response = self.connection.make_request('DELETE', self.name, query_args='bucketqos')
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_requester_qos(self, requester_qos):
        headers = {'content-type': 'application/xml'}
        requester_qos_xml = requester_qos.to_xml()
        if not isinstance(requester_qos_xml, bytes):
            requester_qos_xml = requester_qos_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(requester_qos_xml)
        headers['Content-MD5'] = md5
        logger.debug('bucket: {0}, request_body: {1}'.format(self.name, requester_qos_xml))
        response = self.connection.make_request('PUT', self.name, data=requester_qos_xml, query_args='requesterqos',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_requester_qos(self):
        response = self.connection.make_request('GET', self.name, query_args='requesterqos')
        body = response.read()
        if response.status == 200:
            bucket_quota = RequesterQos(status=response.status, reason=response.reason, headers=response.headers,
                                        raw_body=body)
            h = handler.XmlHandler(bucket_quota, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return bucket_quota
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_requester_qos(self):
        response = self.connection.make_request('DELETE', self.name, query_args='requesterqos')
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_transfer_acceleration(self, enabled):
        bta = BucketTransferAcceleration(enabled=enabled)
        bta_xml = bta.to_xml()
        if not isinstance(bta_xml, bytes):
            bta_xml = bta_xml.encode('utf-8')
        md5 = ks3.utils.compute_base64_md5_digest(bta_xml)
        headers = {'Content-MD5': md5, 'Content-Type': 'application/xml'}
        response = self.connection.make_request('PUT', self.name, data=bta_xml, query_args='transferAcceleration',
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_bucket_transfer_acceleration(self):
        response = self.connection.make_request('GET', self.name, query_args='transferAcceleration')
        body = response.read()
        if response.status == 200:
            bta = BucketTransferAcceleration(status=response.status, reason=response.reason, headers=response.headers,
                                             raw_body=body)
            h = handler.XmlHandler(bta, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return bta
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_bucket_data_redundancy_switch(self, data_redundancy_type=None):
        headers = {'x-kss-data-redundancy-type': data_redundancy_type}

        response = self.connection.make_request('PUT', self.name, headers=headers, query_args='dataRedundancySwitch')
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_object_data_redundancy_transition(self, key="", data_redundancy_type=None):
        headers = {'x-kss-data-redundancy-type': data_redundancy_type}

        response = self.connection.make_request('PUT', self.name, key=key, headers=headers, query_args='dataRedundancyTransition')
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)


    def get_bucket_data_redundancy_switch(self):
        response = self.connection.make_request('GET', self.name, query_args='dataRedundancySwitch')
        body = response.read()
        if response.status == 200:
            bdrs = BucketDataRedundancySwitch(status=response.status, reason=response.reason, headers=response.headers,
                                             raw_body=body)
            h = handler.XmlHandler(bdrs, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return bdrs
        else:
            raise S3ResponseError(response.status, response.reason, body)


    # Public Network Block methods
    def set_public_network_block(self, public_network_block):
        """
        Set the public network block configuration for this bucket

        :type public_network_block: BucketPublicNetworkBlock
        :param public_network_block: The public network block configuration

        :rtype: ResponseResult
        :return: Response result
        """
        return self.connection.set_bucket_public_network_block(self.name, public_network_block)

    def get_public_network_block(self):
        """
        Get the public network block configuration for this bucket

        :rtype: BucketPublicNetworkBlock
        :return: The public network block configuration
        """
        return self.connection.get_bucket_public_network_block(self.name)

    def delete_public_network_block(self):
        """
        Delete the public network block configuration for this bucket

        :rtype: ResponseResult
        :return: Response result
        """
        return self.connection.delete_bucket_public_network_block(self.name)


class BucketLocation(object):
    def __init__(self, *args, **kwargs):
        self.location = ''

        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        pass

    def endElement(self, name, current_text, connection):
        if name == 'LocationConstraint':
            self.location = current_text



# -*- coding: utf-8 -*-
from ks3.connection import Connection
from datetime import datetime

####################公共头#########################

# 金山云主账号 AccessKey 拥有所有API的访问权限，风险很高。强烈建议您创建并使用子账号账号进行 API 访问或日常运维，请登录 https://uc.console.ksyun.com/pro/iam/#/user/list 创建子账号。
# 通过指定 host(Endpoint)，您可以在指定的地域创建新的存储空间。host(Endpoint) 以北京为例，其它 Region 请按实际情况填写。
conn = Connection('<yourAccessKeyId>', '<yourAccessKeySecret>', host='ks3-cn-beijing.ksyuncs.com')


####################内容#########################

# 创建存储空间。
# 如果需要在创建存储空间时设置存储空间访问权限，请参考以下代码。
# 以下以配置存储空间为私有访问权限。
# conn.create_bucket(bucket_name, policy='private')
# 默认创建私有访问权限的存储空间
conn.create_bucket('<yourBucketName>')
# 创建项目 id 为 105150 的归档存储空间
# header 中的 x-kss-bucket-type 为存储空间类型
conn.create_bucket('<yourBucketName>', project_id=105150, headers={"x-kss-bucket-type": "ARCHIVE"})

# 列举所有的存储空间。
# projectIds 为项目id，多个项目用,间隔的
buckets = conn.get_all_buckets(project_ids="25532")
for b in buckets:
	print(b.name)
# 列举存储空间的更多详情，请参见GetService（api地址）

# 获取存储空间的访问权限（暂时无法说明）
# 获取存储空间的访问权限，并打印出ACL设置以及授权的信息
b = conn.get_bucket('<yourBucketName>')
policy = b.get_acl()
for grant in policy.acl.grants:
	print(grant.id, grant.permission)

# 设置存储空间的访问权限
# 获取存储空间实例
b = conn.get_bucket('<yourBucketName>')
# 存储空间的访问权限, private or public-read or public-read-write
b.set_acl("public-read")

# 删除存储空间
conn.delete_bucket('<yourBucketName>')

# 如果存储空间下面存在文件，那么需要首先删除所有文件
for k in b.list():
	k.delete()
conn.delete_bucket('<yourBucketName>')

# 设置存储空间策略
b = conn.get_bucket('<yourBucketName>')
# 存储空间策略，授权id为32432423iam用户对此存储空间有完全控制的权限
policy = '{"Statement":[{"Resource":["krn:ksc:ks3:::jiangran123","krn:ksc:ks3:::jiangran123/*"],"Principal":{"KSC":["krn:ksc:iam::32432423:root"]},"Action":["ks3:*"],"Effect":"Allow"}]}'
# 设置存储空间策略
b.set_bucket_policy(policy=policy)
# 获取存储空间策略
policy = b.get_bucket_policy()
# 删除存储空间策略
b.delete_bucket_policy()

# 抓取网络资源上传
# 从第三方URL拉取文件，并上传至KS3某个 bucket 中存储成名为 object 的文件。
b.fetch_object('www-logo', source_url='http://fe.ksyun.com/project/resource/img/www-logo.png')
# 参见：[金山云官方文档-Put Object Fetch](https://docs.ksyun.com/documents/949)

# 在源存储空间设置复制规则
# 参数为目标桶名称和『是否开启删除复制』
# 若deleteMarkerStatus参数指定为Enabled为开启，若为Disabled或不指定均为关闭状态，若开启删除复制，则当源Bucket删除一个对象时，该对象在目标Bucket的副本也会删除
# 前缀匹配，如果object匹配了前缀规则才会对该对象进行复制，每条复制规则最多添加10条前缀匹配规则，且前缀之间不同重叠
b.set_bucket_crr('<yourTargetBucketName>', deleteMarkerStatus='Disabled', prefix_list=['hello'])
# 获取存储空间复制规则
bucket_crr = b.get_bucket_crr()
print(bucket_crr.to_xml())
# 删除存储空间复制规则
b.delete_bucket_crr()

# 管理生命周期规则
# 查看生命周期规则。
lifecycle = b.get_bucket_lifecycle()
print(lifecycle.to_xml())
# 设置生命周期规则
from ks3.xmlParsers.bucketLifecycle import BucketLifecycle, Rule as LifecycleRule, Filter as LifecycleFilter, Expiration as LifecycleExpiration, Transition as LifecycleTransition
from ks3.tagging import Tag
tag = Tag(key="tag1", value="test1")
# 筛选前缀为 prefix1、标签为 tag1:test1 的 objects，设置过期规则，在其最后修改时间3天后过期。
rule1 = LifecycleRule('rule1', LifecycleFilter('prefix1', tags=[tag]),
                      status='Enabled',
                      expiration=LifecycleExpiration(days=3))

# 设置过期规则，筛选标签为 tag1:test1 的 objects，最后修改时间在指定日期之前的，过期
rule2 = LifecycleRule('rule2', LifecycleFilter(tags=[tag]),
                      status='Enabled',
                      expiration=LifecycleExpiration(
                        date=datetime(2021, 12, 12).isoformat(timespec='seconds') + '+08:00'))

# 设置存储类型转换规则，筛选前缀为 prefix3 的 objects，在其最后修改时间20天之后转为低频访问类型，在其最后修改时间30天之后转为归档类型。
rule3 = LifecycleRule('rule3', LifecycleFilter('prefix3'),
                      status='Enabled',
                      transitions=[LifecycleTransition(days=20, storage_class='STANDARD_IA'),
                                   LifecycleTransition(days=60, storage_class='ARCHIVE')])

# 设置存储类型转换规则，筛选前缀为 prefix3 的 objects，最后修改时间在指定日期之前的，转为低频访问类型
rule4 = LifecycleRule('rule4', LifecycleFilter('prefix4'),
                      status='Enabled',
                      transitions=[
                        LifecycleTransition(date=datetime(2021, 12, 12).isoformat(timespec='seconds') + '+08:00',
                                            storage_class='STANDARD_IA')])
lifecycle = BucketLifecycle([rule1, rule2, rule3, rule4])
b.set_bucket_lifecycle(lifecycle)
# 删除生命周期规则
b.delete_bucket_lifecycle()

# 管理存储空间日志设置功能
# 注：只有bucket的所有者才有权限使用该接口
# 查看日志设置
print(b.get_bucket_logging().to_xml())
# 开启日志设置
from ks3.xmlParsers.bucketLogging import BucketLogging
# 第一个参数为目标桶实例或者目标桶的名称；target_prefix 为日志文件的前缀
b.enable_logging(b, target_prefix='test')
# 或者
# target为目标桶实例或者目标桶的名称；target_prefix 为日志文件的前缀
blogging = BucketLogging(target=b.name, target_prefix='test_log')
print(b.set_bucket_logging(blogging.to_xml()))
# 关闭日志设置
print(b.disable_logging())

# 管理存储空间 CORS 跨域规则
# 查看 CORS 跨域规则
print(b.get_bucket_cors().to_xml())
# 设置 CORS 跨域规则
from ks3.xmlParsers.bucketCors import BucketCors, CORSRule
# origins：允许跨域请求的来源地址；methods：允许跨域的请求方法；max_age: 指定浏览器对特定资源的预取(OPTIONS)请求返回结果的缓存时间,单位为秒; headers: 允许跨域的请求头部；exposed_headers: 允许跨域的响应头部
cors = BucketCors([CORSRule(origins=["http://dev.ksyun.com"], methods=["GET", "HEAD"], max_age="200", headers=["content-type"], exposed_headers=["content-type", "x-kss-acl"])])
# print('cors rules: ', cors.to_xml())
print(b.set_bucket_cors(cors))
# 删除 CORS 跨域规则
b.delete_bucket_cors()

# 管理存储空间回源功能
# 查看回源功能
print(b.get_bucket_mirror())
# 设置回源功能
from ks3.xmlParsers.bucketMirror import HeaderSetting, MirrorRequestSetting, AsyncMirrorRule, SyncMirrorRules, BucketMirror
'''
  异步回源规则，该字段与sync_mirror_rules必须至少有一个，可同时存在。
  mirror_urls: 一组源站url，数量不超过10个，url必须以http或者https开头
  saving_setting_acl: 文件上传KS3时，指定文件的权限。
'''
async_mirror_rule = AsyncMirrorRule.rule_with_acl(mirror_urls=["http://abc.om", "http://www.wps.cn"], saving_setting_acl="private")
# 自定义header，这些header的key和value均是固定的，ks3请求源站时会带上这些header。
set_headers = [{
    "key": "d",
    "value": "b"
}]
# 从客户端发给ks3的header中移除以下指定的header，通常与pass_all或者pass_headers配合使用，只能指定header中的key，不能指定value。
remove_headers = [{
    "key": "d"
}]
# 将客户端发给ks3的header中指定的几个透传给源站，只能指定header中的key，不能指定value。
pass_headers = [{
    "key": "abc"
}]

# pass_all: 将客户端发给ks3的header全部透传给源站，该字段与pass_headers互斥。
header_setting = HeaderSetting(set_headers=set_headers, remove_headers=remove_headers, pass_all=False, pass_headers=pass_headers)

# pass_query_string: ks3请求源站时是否将客户端请求ks3时的query string透传给源站。
# follow3xx: 设置访问源站时，是否follow 302/301。ks3是否响应源站的301和302跳转，如果为false且源站返回了302，则ks3会返回424给客户端，如果为true则ks3收到302后会请求302的location。
mirror_request_setting = MirrorRequestSetting(pass_query_string=False, follow3xx=False, header_setting=header_setting)
sync_mirror_rules = SyncMirrorRules.rules_with_prefix_acl(key_prefixes=["abc"], mirror_url="http://v-ks-a-i.originalvod.com", mirror_request_setting=mirror_request_setting, saving_setting_acl="private")

# use_default_robots: 是否使用默认的robots.txt，如果为true则会在bucket下生成一个robots.txt。
mirror = BucketMirror(use_default_robots=False, async_mirror_rule=async_mirror_rule, sync_mirror_rules=[sync_mirror_rules])
print(b.set_bucket_mirror(mirror))
# 删除回源功能
b.delete_bucket_mirror()





import logging
import os

import ks3.utils
from ks3 import user
from ks3 import handler
from ks3.exception import KS3ClientError
from ks3.responseResult import ResponseMetadata, ResponseResult, IterableResponseResult

try:
    from ks3.encryption import Crypts
except:
    pass
import xml.sax


logger = logging.getLogger(__name__)


class CompleteMultiPartUpload(object):
    """
    Represents a completed MultiPart Upload.  Contains the
    following useful attributes:

     * location - The URI of the completed upload
     * bucket_name - The name of the bucket in which the upload
                     is contained
     * key_name - The name of the new, completed key
     * etag - The MD5 hash of the completed, combined upload
     * version_id - The version_id of the completed upload
     * encrypted - The value of the encryption header
    """

    def __init__(self, bucket=None, *args, **kwargs):
        self.bucket = bucket
        self.location = None
        self.bucket_name = None
        self.key_name = None
        self.etag = None
        self.version_id = None
        self.encrypted = None
        self.status = None

        self.response_metadata = ResponseMetadata(**kwargs)

    def __repr__(self):
        return '<CompleteMultiPartUpload: %s.%s>' % (self.bucket_name,
                                                     self.key_name)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name.startswith("ns2"):
            name = name.split(":")[1]
        if name == 'Location':
            self.location = value
        elif name == 'Bucket':
            self.bucket_name = value
        elif name == 'Key':
            self.key_name = value
        elif name == 'ETag':
            self.etag = value
        else:
            setattr(self, name, value)


class Part(object):
    """
    Represents a single part in a MultiPart upload.
    Attributes include:

     * part_number - The integer part number
     * last_modified - The last modified date of this part
     * etag - The MD5 hash of this part
     * size - The size, in bytes, of this part
    """

    def __init__(self, bucket=None):
        self.bucket = bucket
        self.part_number = None
        self.last_modified = None
        self.etag = None
        self.size = None
        self.part_crc = None

    def __repr__(self):
        if isinstance(self.part_number, int):
            return '<Part %d>' % self.part_number
        else:
            return '<Part %s>' % None

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name.startswith("ns2"):
            name = name.split(":")[1]
        if name == 'PartNumber':
            self.part_number = int(value)
        elif name == 'LastModified':
            self.last_modified = value
        elif name == 'ETag':
            self.etag = value
        elif name == 'Size':
            self.size = int(value)
        else:
            setattr(self, name, value)


def part_lister(mpupload, part_number_marker=None):
    """
    A generator function for listing parts of a multipart upload.
    """
    more_results = True
    part = None
    while more_results:
        parts = mpupload.get_all_parts(None, part_number_marker)
        for part in parts:
            yield part
        part_number_marker = mpupload.next_part_number_marker
        more_results = mpupload.is_truncated


class MultiPartUpload(object):
    """
    Represents a MultiPart Upload operation.
    """

    def __init__(self, bucket=None, *args, **kwargs):
        self.bucket = bucket
        self.bucket_name = None
        self.key_name = None
        self.id = id
        self.initiator = None
        self.owner = None
        self.storage_class = None
        self.initiated = None
        self.part_number_marker = None
        self.next_part_number_marker = None
        self.max_parts = None
        self.is_truncated = False
        self._parts = None
        self.crypt_context = None
        # self.size_accumulator = 0
        self.size_accumulator = {}
        self.total_size = 0
        self.part_crc_infos = {}

        self.response_metadata = ResponseMetadata(**kwargs)


    def __repr__(self):
        return '<MultiPartUpload %s>' % self.key_name

    def __iter__(self):
        return part_lister(self)

    def to_xml(self):
        s = '<CompleteMultipartUpload>\n'
        for part in self:
            s += '  <Part>\n'
            s += '    <PartNumber>%d</PartNumber>\n' % part.part_number
            s += '    <ETag>%s</ETag>\n' % part.etag
            s += '  </Part>\n'
        s += '</CompleteMultipartUpload>'
        return s

    def set_crypt_context(self, crypt):
        self.crypt_context = crypt

    def startElement(self, name, attrs, connection):
        if name.startswith("ns2"):
            name = name.split(":")[1]
        if name == 'Initiator':
        # if name == 'ns2:Initiator':
            self.initiator = user.User(self)
            return self.initiator
        elif name == 'Owner':
        # elif name == 'ns2:Owner':
            self.owner = user.User(self)
            return self.owner
        elif name == 'Part':
        # elif name == 'ns2:Part':
            part = Part(self.bucket)
            self._parts.append(part)
            return part
        return None

    def endElement(self, name, value, connection):
        if name.startswith("ns2"):
            name = name.split(":")[1]
        if name == 'Bucket':
            self.bucket_name = value
        elif name == 'Key':
            self.key_name = value
        elif name == 'UploadId':
            self.id = value
        elif name == 'StorageClass':
            self.storage_class = value
        elif name == 'PartNumberMarker':
            self.part_number_marker = value
        elif name == 'NextPartNumberMarker':
            self.next_part_number_marker = value
        elif name == 'MaxParts':
            self.max_parts = int(value)
        elif name == 'IsTruncated':
            if value == 'true':
                self.is_truncated = True
            else:
                self.is_truncated = False
        elif name == 'Initiated':
            self.initiated = value
        else:
            setattr(self, name, value)

    def get_all_parts(self, max_parts=None, part_number_marker=None,
                      encoding_type=None):
        """
        Return the uploaded parts of this MultiPart Upload.  This is
        a lower-level method that requires you to manually page through
        results.  To simplify this process, you can just use the
        object itself as an iterator and it will automatically handle
        all of the paging with S3.
        """
        result = self.bucket.list_parts(
            bucket_name=self.bucket.name,
            key_name=self.key_name,
            upload_id=self.id,
            max_parts=max_parts,
            part_number_marker=part_number_marker,
            encoding_type=encoding_type
        )
        self._parts = result.data
        self.part_number_marker = part_number_marker
        return result

    def upload_part_from_file(self, fp, part_num, headers=None, replace=True,
                              cb=None, num_cb=10, md5=None, size=None, is_last_part="origin"):
        """
        Upload another part of this MultiPart Upload.

        .. note::

            After you initiate multipart upload and upload one or more parts,
            you must either complete or abort multipart upload in order to stop
            getting charged for storage of the uploaded parts. Only after you
            either complete or abort multipart upload, Amazon S3 frees up the
            parts storage and stops charging you for the parts storage.

        :type fp: file
        :param fp: The file object you want to upload.

        :type part_num: int
        :param part_num: The number of this part.

        The other parameters are exactly as defined for the
        :class:`boto.s3.key.Key` set_contents_from_file method.

        :rtype: :class:`boto.s3.key.Key` or subclass
        :returns: The uploaded part containing the etag.
        """
        if part_num < 1:
            raise ValueError('Part numbers must be greater than zero')

        # For local encryption, we need to calculate some values before calling bucket method
        if self.bucket.connection.local_encrypt and self.crypt_context:
            # Get sizes of the whole file and the part
            fp.seek(0, os.SEEK_END)
            part_size = fp.tell()
            fp.seek(0, os.SEEK_SET)
            assert part_size, "upload part size can not be 0 !"
            self.total_size = os.fstat(fp.fileno()).st_size

            # 看起来这一句没必要，如果相等，is_last_part肯定是True
            if part_size == self.total_size:
                assert is_last_part != "origin", "Please indicate 'is_last_part=False/True' \
                                                  when calling upload_part_from_file."
            else:
                size_now = self.add_all_part(part_num)
                temp_accumulator = size_now + part_size
                if temp_accumulator == self.total_size:
                    is_last_part = True
                else:
                    is_last_part = False

            if not is_last_part:
                assert (part_size % 16 == 0), "The part size must be multiples of 16 except the last part in local encrypt mode."

            # Set crypt context properties
            self.crypt_context.part_num = part_num
            self.crypt_context.is_last_part = is_last_part
            self.crypt_context.action_info = "upload_part"

            # Call bucket method
            response = self.bucket.upload_part_from_file(
                key_name=self.key_name,
                upload_id=self.id,
                fp=fp,
                part_num=part_num,
                headers=headers,
                replace=replace,
                cb=cb,
                num_cb=num_cb,
                md5=md5,
                size=size,
                crypt_context=self.crypt_context,
                is_last_part=is_last_part
            )

            # key.set_contents_from_file() raise an exception when fail.
            self.size_accumulator[part_num] = part_size
        else:
            # Call bucket method
            response = self.bucket.upload_part_from_file(
                key_name=self.key_name,
                upload_id=self.id,
                fp=fp,
                part_num=part_num,
                headers=headers,
                replace=replace,
                cb=cb,
                num_cb=num_cb,
                md5=md5,
                size=size,
                crypt_context=None,
                is_last_part=is_last_part
            )

        # Handle CRC if enabled
        if self.bucket.connection.enable_crc:
            part_size = fp.tell()
            if self.bucket.connection.local_encrypt and self.crypt_context:
                if is_last_part is True:
                    # After local encryption, the size of the last part is filled to a multiple of 16
                    pad = 16 - part_size % 16
                    part_size += pad
                elif part_num == 1:
                    # Local encryption sets the key ciphertext with a size of 16 bytes in the first part of the file
                    part_size += 16
            part_info = Part()
            part_info.size = part_size
            part_info.part_crc = response.getheader(self.bucket.connection.provider.checksum_crc64ecma_header)
            self.part_crc_infos[part_num] = part_info

        return response

    def add_all_part(self, part_num):
        sum = 0
        for i in range(1,part_num):
            if self.size_accumulator.get(i):
                sum += self.size_accumulator[i]
        return sum

    def copy_part_from_key(self, src_bucket_name, src_key_name, part_num,
                           start=None, end=None,
                           headers=None):
        """
        Copy another part of this MultiPart Upload.

        :type src_bucket_name: string
        :param src_bucket_name: Name of the bucket containing the source key

        :type src_key_name: string
        :param src_key_name: Name of the source key

        :type part_num: int
        :param part_num: The number of this part.

        :type start: int
        :param start: Zero-based byte offset to start copying from

        :type end: int
        :param end: Zero-based byte offset to copy to

        :type headers: dict
        :param headers: Any headers to pass along in the request
        """
        return self.bucket.copy_part_from_key(
            key_name=self.key_name,
            upload_id=self.id,
            src_bucket_name=src_bucket_name,
            src_key_name=src_key_name,
            part_num=part_num,
            start=start,
            end=end,
            headers=headers
        )

    def complete_upload(self, headers=None):
        """
        Complete the MultiPart Upload operation.  This method should
        be called when all parts of the file have been successfully
        uploaded to S3.

        :rtype: :class:`boto.s3.multipart.CompletedMultiPartUpload`
        :returns: An object representing the completed upload.
        """
        xml = self.to_xml()
        resp = self.bucket.complete_multipart_upload(self.key_name,
                                                     self.id, xml, headers)
        if self.bucket.connection.enable_crc and len(self.part_crc_infos) > 0:
            # Iterating from 0 is equivalent to iterating from the first crc value
            parts_crc = 0
            crc_obj = ks3.utils.Crc64(0)

            part_tuples = sorted(self.part_crc_infos.items(), key=lambda i: i[0])
            for t in part_tuples:
                p = t[1]
                parts_crc = crc_obj.combine(parts_crc, int(p.part_crc), p.size)
                logger.debug('p: {0} part_crc: {1}, size: {2}, combined crc: {3}'
                             .format(p, p.part_crc, p.size, parts_crc))
            logger.debug('upload key: {0}, bucket: {1}, check crc: [client crc: {2}, server crc: {3}]'
                         .format(self.key_name, self.bucket.name, parts_crc, resp.ChecksumCRC64ECMA))
            if not ks3.utils.check_crc(str(parts_crc), resp.ChecksumCRC64ECMA):
                raise KS3ClientError("Inconsistent CRC checksum, "
                                     "client_crc: %s, server_crc: %s" % (str(parts_crc), resp.ChecksumCRC64ECMA),
                                     resp.response_metadata.headers['x-kss-request-id'])
        return resp

    def cancel_upload(self):
        """
        Cancels a MultiPart Upload operation.  The storage consumed by
        any previously uploaded parts will be freed. However, if any
        part uploads are currently in progress, those part uploads
        might or might not succeed. As a result, it might be necessary
        to abort a given multipart upload multiple times in order to
        completely free all storage consumed by all parts.
        """
        return self.bucket.cancel_multipart_upload(self.key_name, self.id)


PartInfo = Part



from ks3.responseResult import ResponseMetadata

'''
block_type:
  阻止公网访问类型
  All：阻止所有公网访问
  ExcludeAuthorization：阻止公网访问，除有效鉴权
  ExcludeConsole：阻止公网访问，控制台除外
'''
class PublicNetworkBlock:
    def __init__(self, block_type=None, *args, **kwargs):
        self.block_type = block_type
        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'BlockType':
            self.block_type = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = u'<PublicNetworkBlockConfiguration>'
        if self.block_type is not None:
            s += '<BlockType>{0}</BlockType>'.format(self.block_type)
        s += '</PublicNetworkBlockConfiguration>'
        return s


class BucketPublicNetworkBlock:
    def __init__(self, block_type=None, *args, **kwargs):
        self.block_type = block_type
        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'BlockType':
            self.block_type = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = u'<BucketPublicNetworkBlockConfiguration>'
        if self.block_type is not None:
            s += '<BlockType>{0}</BlockType>'.format(self.block_type)
        s += '</BucketPublicNetworkBlockConfiguration>'
        return s


from ks3.responseResult import ResponseMetadata


class MigrationConfiguration:
    def __init__(self, migration_id=None, status=None, migration_failure=None, operation=None, creation_time=None,
                 termination_time=None, *args, **kwargs):
        self.migration_id = migration_id
        self.status = status
        self.migration_failure = migration_failure
        self.operation = operation
        self.creation_time = creation_time
        self.termination_time = termination_time

        self.response_metadata = ResponseMetadata(**kwargs)

    def to_xml(self):
        s = '<MigrationConfiguration>'
        if self.migration_id is not None:
            s += '<Id>%s</Id>' % self.migration_id
        if self.status is not None:
            s += '<Status>%s</Status>' % self.status
        if self.migration_failure is not None:
            s += self.migration_failure.to_xml()
        if self.operation is not None:
            s += self.operation.to_xml()
        if self.creation_time is not None:
            s += '<CreationTime>%s</CreationTime>' % self.creation_time
        if self.termination_time is not None:
            s += '<TerminationTime>%s</TerminationTime>' % self.termination_time
        s += '</MigrationConfiguration>'
        return s

    def startElement(self, name, attrs, connection):
        if name == 'MigrationFailure':
            self.migration_failure = MigrationFailure()
            return self.migration_failure
        elif name == 'Operation':
            self.operation = Operation()
            return self.operation
        else:
            return None

    def endElement(self, name, value, connection):
        if name == 'Id':
            self.migration_id = value
        elif name == 'Status':
            self.status = value
        elif name == 'CreationTime':
            self.creation_time = value
        elif name == 'TerminationTime':
            self.termination_time = value
        else:
            setattr(self, name, value)


class MigrationFailure:
    def __init__(self, code=None, reason=None):
        self.code = code
        self.reason = reason

    def to_xml(self):
        s = '<MigrationFailure>'
        if self.code is not None:
            s += '<Code>%s</Code>' % self.code
        if self.reason is not None:
            s += '<Reason>%s</Reason>' % self.reason
        s += '</MigrationFailure>'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Code':
            self.code = value
        elif name == 'Reason':
            self.reason = value
        else:
            setattr(self, name, value)


class Operation:
    def __init__(self, migration_source=None, migration_dest=None, storage_class=None):
        self.migration_source = migration_source
        self.migration_dest = migration_dest
        self.storage_class = storage_class

    def to_xml(self):
        s = '<Operation>'
        if self.migration_source is not None:
            s += '<MigrationSource>%s</MigrationSource>' % self.migration_source
        if self.migration_dest is not None:
            s += '<MigrationDest>%s</MigrationDest>' % self.migration_dest
        if self.storage_class is not None:
            s += '<StorageClass>%s</StorageClass>' % self.storage_class
        s += '</Operation>'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'MigrationSource':
            self.migration_source = value
        elif name == 'MigrationDest':
            self.migration_dest = value
        elif name == 'StorageClass':
            self.storage_class = value
        else:
            setattr(self, name, value)


class MigrationResult:
    def __init__(self, migration_id=None, *args, **kwargs):
        self.migration_id = migration_id
        self.response_metadata = ResponseMetadata(**kwargs)

    def to_xml(self):
        s = '<MigrationResult>'
        if self.migration_id is not None:
            s += '<Id>%s</Id>' % self.migration_id
        s += '</MigrationResult>'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Id':
            self.migration_id = value
        else:
            setattr(self, name, value)


# -*- coding: utf-8 -*-
import os

from ks3.connection import Connection
from ks3.xmlParsers.publicNetworkBlock import PublicNetworkBlock, BucketPublicNetworkBlock

# 首先初始化AccessKeyId、AccessKeySecret、Endpoint等信息。
# 通过环境变量获取，或者把诸如"<你的AccessKeyId>"替换成真实的AccessKeyId等。
ak = os.getenv('KS3_IAM_ACCESS_KEY_ID', '<YOUR_ACCESS_KEY>')
sk = os.getenv('KS3_IAM_ACCESS_KEY_SECRET', '<YOUR_SECRET_KEY>')
bucket_name = os.getenv('KS3_BUCKET', '<KS3_TEST_BUCKET>')
endpoint = 'ks3-cn-beijing.ksyuncs.com'

conn = Connection(ak, sk, host=endpoint, is_secure=False)

def set_global_public_network_block():
    """设置全局公共网络访问阻止策略"""
    try:
        # 创建公共网络阻止配置
        pnb = PublicNetworkBlock(block_type='All')

        # 设置全局公共网络访问阻止
        resp = conn.set_public_network_block(pnb)
        print("设置全局公共网络访问阻止成功，request_id:", resp.response_metadata.request_id)
    except Exception as e:
        print("设置全局公共网络访问阻止失败:", e)
        if hasattr(e, 'request_id'):
            print("request_id:", e.request_id)

def get_global_public_network_block():
    """获取全局公共网络访问阻止策略"""
    try:
        # 获取全局公共网络访问阻止配置
        pnb = conn.get_public_network_block()
        print("获取全局公共网络访问阻止成功:")
        print("BlockType:", pnb.block_type)
        print("request_id:", pnb.response_metadata.request_id)
    except Exception as e:
        print("获取全局公共网络访问阻止失败:", e)
        if hasattr(e, 'request_id'):
            print("request_id:", e.request_id)

def delete_global_public_network_block():
    """删除全局公共网络访问阻止策略"""
    try:
        # 删除全局公共网络访问阻止配置
        resp = conn.delete_public_network_block()
        print("删除全局公共网络访问阻止成功，request_id:", resp.response_metadata.request_id)
    except Exception as e:
        print("删除全局公共网络访问阻止失败:", e)
        if hasattr(e, 'request_id'):
            print("request_id:", e.request_id)

def set_bucket_public_network_block():
    """设置Bucket级别的公共网络访问阻止策略"""
    try:
        # 获取Bucket对象
        bucket = conn.get_bucket(bucket_name)

        # 创建Bucket公共网络阻止配置
        bpnb = BucketPublicNetworkBlock(block_type='0')

        # 设置Bucket公共网络访问阻止
        resp = bucket.set_public_network_block(bpnb)
        print("设置Bucket公共网络访问阻止成功，request_id:", resp.response_metadata.request_id)
    except Exception as e:
        print("设置Bucket公共网络访问阻止失败:", e)
        if hasattr(e, 'request_id'):
            print("request_id:", e.request_id)

def get_bucket_public_network_block():
    """获取Bucket级别的公共网络访问阻止策略"""
    try:
        # 获取Bucket对象
        bucket = conn.get_bucket(bucket_name)

        # 获取Bucket公共网络访问阻止配置
        bpnb = bucket.get_public_network_block()
        print("获取Bucket公共网络访问阻止成功:")
        print("BlockType:", bpnb.block_type)
        print("request_id:", bpnb.response_metadata.request_id)
    except Exception as e:
        print("获取Bucket公共网络访问阻止失败:", e)
        if hasattr(e, 'request_id'):
            print("request_id:", e.request_id)

def delete_bucket_public_network_block():
    """删除Bucket级别的公共网络访问阻止策略"""
    try:
        # 获取Bucket对象
        bucket = conn.get_bucket(bucket_name)

        # 删除Bucket公共网络访问阻止配置
        resp = bucket.delete_public_network_block()
        print("删除Bucket公共网络访问阻止成功，request_id:", resp.response_metadata.request_id)
    except Exception as e:
        print("删除Bucket公共网络访问阻止失败:", e)
        if hasattr(e, 'request_id'):
            print("request_id:", e.request_id)

if __name__ == "__main__":
    print("=== Public Network Block 示例 ===")

    # 演示全局公共网络访问阻止功能
    print("\n1. 设置全局公共网络访问阻止:")
    set_global_public_network_block()

    print("\n2. 获取全局公共网络访问阻止:")
    get_global_public_network_block()

    print("\n3. 删除全局公共网络访问阻止:")
    delete_global_public_network_block()

    # 演示Bucket级别的公共网络访问阻止功能
    print("\n4. 设置Bucket公共网络访问阻止:")
    set_bucket_public_network_block()

    print("\n5. 获取Bucket公共网络访问阻止:")
    get_bucket_public_network_block()

    print("\n6. 删除Bucket公共网络访问阻止:")
    delete_bucket_public_network_block()

    print("\n=== 示例结束 ===")


from ks3.responseResult import ResponseMetadata


class BucketVersioningConfig(object):
    def __init__(self, status=None, *args, **kwargs):
        self.status = status

        self.response_metadata = ResponseMetadata(**kwargs)

    def __repr__(self):
        return "status " + self.status

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Status':
            self.status = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<?xml version="1.0" encoding="UTF-8"?>'
        s += '<VersioningConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
        if self.status is not None:
            s += '<Status>%s</Status>' % self.status
        s += '</VersioningConfiguration>'
        return s



from ks3.responseResult import ResponseMetadata

"""
<AccessMonitorConfiguration>
  <Status>Enabled</Status>
</AccessMonitorConfiguration>
"""


class BucketAccessMonitor(object):

    def __init__(self, status=None, *args, **kwargs):
        """
        :param status: Enabled 或者 Disabled
        """
        self.status = status

        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, value, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Status':
            self.status = value

    def to_xml(self):
        s = '<AccessMonitorConfiguration>' \

        if self.status is not None:
            s += '<Status>%s</Status>' % self.status
        s += '</AccessMonitorConfiguration>'
        return s



# -*- coding: utf-8 -*-
from ks3.sts import assumeRole
from datetime import datetime

####################公共头#########################

# 金山云主账号 AccessKey 拥有所有API的访问权限，风险很高。强烈建议您创建并使用子账号账号进行 API 访问或日常运维，请登录 https://uc.console.ksyun.com/pro/iam/#/user/list 创建子账号。
# bucket_names 为一个或者多个 bucket 名称，逗号间隔
# start_time 和 end_time 分别为起始时间和结束时间，时间格式为%Y%m%d%H%M%S，比如202111190000
# products 指的是查询单个或者多个计费项名称，逗号间隔；比如DataSize,RequestsGet；如果不填，则查询所有计费项（除带宽，带宽需单独查询）
data = assumeRole('<yourAccessKeyId>', '<yourAccessKeySecret>', "krn:ksc:iam::xxx:role/xx-test-bucket", "ks3")
print(data)




import os
import logging
from Crypto.Cipher import AES
from Crypto.Protocol.KDF import PBKDF2
from Crypto.Util.Padding import pad, unpad
from Crypto.Hash import SHA256
import base64
import ks3
logger = logging.getLogger(__name__)

PBKDF2_ITER = 10000
PBKDF2_KEY_LENGTH = 32
TOKEN_SEPARATOR = "_"
IV = 'ks3-share12345v1'


def parse_token(token, access_code):
    result_arr = token.split(TOKEN_SEPARATOR)
    code = result_arr[0]
    logger.debug(f"token: {token}")
    salt = base64.b64decode(result_arr[1])
    iv = base64.b64decode(result_arr[2])
    logger.debug(f"encrypted_b64: {code}, salt_b64: {result_arr[1]}, iv_b64: {result_arr[2]}")
    key = PBKDF2(access_code, salt, dkLen=PBKDF2_KEY_LENGTH, count=PBKDF2_ITER, hmac_hash_module=SHA256)
    logger.debug(f"PBKDF2 key: {key.hex()}, iv: {iv}")
    return {
        'authCode': code,
        'key': key,
        'iv': iv
    }


class ShareCryptor(object):
    def __init__(self, access_code):
        """
        This encryption class use AES.CBC MODE(128 bit) as default.
        @param access_code: 分享码
        """
        self.access_code = access_code

    def decrypt(self, token_params):
        logger.debug('\ndecrypting...')
        # share_token = token_params[9:]
        share_token = token_params
        token_params = parse_token(share_token, self.access_code)
        if not token_params:
            return ''
        share_token = token_params['authCode']
        key = token_params['key']
        iv = token_params['iv']

        try:
            cipher = AES.new(key, AES.MODE_CBC, iv=iv)
            decoded_token = base64.b64decode(share_token)
            logger.debug(f'encrypted: ${decoded_token.hex()}')
            decrypted = cipher.decrypt(decoded_token)
            logger.debug(f'padded_content: ${decrypted}')
            unpad_content = unpad(decrypted, AES.block_size, style='pkcs7').decode('utf-8')
            logger.debug(f'unpad_content: ${unpad_content}')
            url = base64.b64decode(unpad_content).decode('utf-8')
        except Exception as e:
            raise ValueError("Decryption failed")

        return url

    def encrypt(self, value, salt=None, iv=None):
        logger.debug('encrypting...')
        logger.debug(f'content before base64: {value}')
        content = base64.b64encode(value.encode('utf-8')).decode('utf-8')
        if salt is None:
            salt = b'ks3-share12345v1'
        if iv is None:
            iv = IV.encode('utf-8')
        logger.debug(f"access_code: {self.access_code}, content: {content}, salt: {salt}, iv: {iv}")
        key = PBKDF2(self.access_code, salt, dkLen=PBKDF2_KEY_LENGTH, count=PBKDF2_ITER, hmac_hash_module=SHA256)
        logger.debug(f"PBKDF2 key: {key.hex()}")
        cipher = AES.new(key, AES.MODE_CBC, iv=iv)
        padded_content = pad(content.encode('utf-8'), AES.block_size, style='pkcs7')
        logger.debug(f"padded_content: {padded_content}")
        encrypted = cipher.encrypt(padded_content)
        logger.debug(f"encrypted: {encrypted.hex()}")
        encrypted_b64 = base64.b64encode(encrypted).decode('utf-8')
        salt_b64 = base64.b64encode(salt).decode('utf-8')
        iv_b64 = base64.b64encode(iv).decode('utf-8')
        logger.debug(f"encrypted_b64: {encrypted_b64}, salt_b64: {salt_b64}, iv_b64: {iv_b64}")
        return f"{encrypted_b64}_{salt_b64}_{iv_b64}"



#!/usr/bin/env python
#! usr/bin/python #coding=utf-8
from Crypto.Cipher import AES
from Crypto import Random
import time

class Crypts(object):
    """
    This encryption class use AES.CBC MODE(128 bit) as default. 
    @param key: your private key. We don't save user's key, so keep it safe or your data will never be decrypted :(
    """
    def __init__(self, key):
        self.key = key
        self.first_iv = Random.new().read(AES.block_size)
        self.calc_iv = ""
        self.block_size = 16
        self.pad_num = 0
        self.part_num = 0
        self.iv_dict = {}
        self.pad_dict = self.init_pad_dict()
        self.calc_md5 = True
        self.action_info = ""
        self.is_last_part = False

    def init_pad_dict(self):
        temp_dict={}
        gap = ord('a')-10
        for i in range(1, 10):
            temp_dict[i] = str(i)
        for i in range(10, 17):
            temp_dict[i] = chr(gap+i)
        return temp_dict

    def set_calc_iv(self,calc_iv):
        self.calc_iv = calc_iv

    def encrypt_without_padding(self, content, iv):
        """
        For those part those size are definitly multiple of 16, use this function. If you need data padding, use the encrypt() func.
        """
        cryptor = AES.new(self.key,AES.MODE_CBC,iv)
        return cryptor.encrypt(content) 

    def encrypt(self, content, iv):
        cryptor = AES.new(self.key,AES.MODE_CBC,iv)
        pad_num = self.block_size-(len(content)%(self.block_size))
        pad_content_char = self.pad_dict[pad_num]
        pad = lambda s: s + ((AES.block_size - len(s) % AES.block_size) * pad_content_char).encode()
        self.pad_num = pad_num
        #identifier + content may still not a multiple of 16, add extra '0' after the first pad
        return cryptor.encrypt(pad(content))

    def decrypt(self, content, iv):
        cryptor = AES.new(self.key,AES.MODE_CBC,iv)
        full_content = cryptor.decrypt(content)
        return full_content

    @staticmethod
    def generate_key(path, file_name):
        sk = Random.new().read(16)
        f = open(path+"/"+file_name,"wb")
        f.write(sk)
        f.close()
    
if __name__ == '__main__':
    # cry = Crypts.generate_key('./', 'try_key')
    cry = Crypts("1234567890123456")
    str0 = ""
    # str0 = "16161616161616161616161616161616"
    # e0 = cry.encrypt(str0,cry.first_iv)
    # print "e0:"+cry.decrypt(e0,cry.first_iv)
    # str1 = "123456789012345"
    # str2 = "sdf1234565432121"
    # str_all="1234567890123456sdf1234565432123"
    # e1 = cry.encrypt(str1,cry.first_iv)
    # e2 = cry.encrypt(str2,e1)
    # e3 = cry.encrypt(str_all,cry.first_iv)
    #
    #
    # r1=cry.decrypt(e1,cry.first_iv)
    # print r1,len(r1)
    # print cry.decrypt(e2,e1)
    # print cry.decrypt(e3,cry.first_iv)
    
